/***************************************************************************
 * Gens: Lens Flare. (S2HD)                                                *
 *                                                                         *
 * Copyright (c) 2009 by David Korth                                       *
 * Images provided by Sonic65.                                             *
 *                                                                         *
 * This program is free software; you can redistribute it and/or modify it *
 * under the terms of the GNU General Public License as published by the   *
 * Free Software Foundation; either version 2 of the LicensE, or (at your  *
 * option) any later version.                                              *
 *                                                                         *
 * This program is distributed in the hope that it will be useful, but     *
 * WITHOUT ANY WARRANTY; without even the implied warranty of              *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License along *
 * with this program; if not, write to the Free Software Foundation, Inc., *
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.           *
 ***************************************************************************/

#include "lens_flare.hpp"
#include "lens_flare_img.h"

#include <stdint.h>
#include <stdlib.h>

#include "gens_core/vdp/vdp_rend.h"
#include "v_inline.h"

#include <png.h>

// Lens Flare variables.
static bool lf_init = false;
static uint8_t	lfraw_320_224[320*224*4];
static uint8_t	lfraw_256_224[320*224*4];


// PNG-related variables and functions.
#include <png.h>

static const uint8_t	*lens_png_dataptr;
static unsigned int	lens_png_datalen;
static unsigned int	lens_png_datapos;


/**
 * lens_png_user_read_data(): libpng user-specified read data function.
 * Used to read PNG data from memory instead of from a file.
 * @param png_ptr Pointer to the PNG information struct.
 * @param png_bytep Pointer to memory to write PNG data to.
 * @param length Length of data requested.
 */
static void lens_png_user_read_data(png_structp png_ptr, png_bytep data, png_size_t length)
{
	// Mark unused parameters.
	((void)png_ptr);
	
	// Make sure there's enough data available.
	if (lens_png_datapos + length > lens_png_datalen)
	{
		// Not enough data is available.
		// TODO: This may still result in a crash. Use longjmp()?
		
		// Zero the buffer.
		memset(data, 0x00, length);
		
		// Return the rest of the buffer.
		length = lens_png_datalen - lens_png_datapos;
		if (length <= 0)
			return;
	}
	
	// Copy the data.
	memcpy(data, &lens_png_dataptr[lens_png_datapos], length);
	
	// Increment the data position.
	lens_png_datapos += length;
}


/**
 * lens_read_png(): Read a PNG image.
 * @param png_data PNG image data.
 * @param png_len Length of the PNG image data.
 * @param w Width.
 * @param h Height.
 * @param out Output buffer.
 * @return 0 on success; non-zero on error.
 */
static int lens_read_png(const uint8_t *png_data, unsigned int png_len,
			 unsigned int w, unsigned int h,
			 uint8_t *out)
{
	// Initialize libpng.
	png_structp png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
	if (!png_ptr)
		return 1;
	
	png_infop info_ptr = png_create_info_struct(png_ptr);
	if (!info_ptr)
	{
		png_destroy_read_struct(&png_ptr, NULL, NULL);
		return 2;
	}
	
	png_infop end_info = png_create_info_struct(png_ptr);
	if (!end_info)
	{
		png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
		return 3;
	}
	
	if (setjmp(png_jmpbuf(png_ptr)))
	{
		// An error occurred while attepmting to decode the PNG image.
		png_destroy_read_struct(&png_ptr, &info_ptr, &end_info);
		return 4;
	}
	
	// Initialize the custom read function.
	lens_png_dataptr = png_data;
	lens_png_datalen = png_len;
	lens_png_datapos = 0;
	
	void *read_io_ptr = png_get_io_ptr(png_ptr);
	png_set_read_fn(png_ptr, read_io_ptr, &lens_png_user_read_data);
	
	// Get the PNG information.
	png_read_info(png_ptr, info_ptr);
	
	// Get the PNG information.
	png_uint_32 width, height;
	int bit_depth, color_type, interlace_type, compression_type, filter_method;
	bool has_alpha = false;
	
	png_get_IHDR(png_ptr, info_ptr, &width, &height, &bit_depth, &color_type,
		     &interlace_type, &compression_type, &filter_method);
	
	if (width != w || height != h)
	{
		// Not the correct size..
		png_destroy_read_struct(&png_ptr, &info_ptr, &end_info);
		return 5;
	}
	
	// Make sure RGB color is used.
	if (color_type == PNG_COLOR_TYPE_PALETTE)
		png_set_palette_to_rgb(png_ptr);
	else if (color_type == PNG_COLOR_TYPE_GRAY)
		png_set_gray_to_rgb(png_ptr);
	else if (color_type == PNG_COLOR_TYPE_RGB_ALPHA)
		has_alpha = true;
	
	// GTK+ expects RGBA format.
	// TODO: Check if this is the same on big-endian machines.
	
	// Convert tRNS to alpha channel.
	if (png_get_valid(png_ptr, info_ptr, PNG_INFO_tRNS))
	{
		png_set_tRNS_to_alpha(png_ptr);
		has_alpha = true;
	}
	
	// Convert 16-bit per channel PNG to 8-bit.
	if (bit_depth == 16)
		png_set_strip_16(png_ptr);
	
	// Get the new PNG information.
	png_get_IHDR(png_ptr, info_ptr, &width, &height, &bit_depth, &color_type,
		     &interlace_type, &compression_type, &filter_method);
	
	// Check if the PNG image has an alpha channel.
	if (!has_alpha)
	{
		// No alpha channel specified.
		// Use filler instead.
		png_set_filler(png_ptr, 0xFF, PNG_FILLER_AFTER);
	}
	
	// Update the PNG info.
	png_read_update_info(png_ptr, info_ptr);
	
	// Create the row pointers.
	png_bytep *row_pointers = (png_bytep*)malloc(h * sizeof(png_bytep));
	for (unsigned int i = 0; i < h; i++)
	{
		row_pointers[i] = out;
		out += (w * 4);
	}
	
	// Read the image data.
	png_read_image(png_ptr, row_pointers);
	
	// Free the row pointers.
	free(row_pointers);
	
	// Close the PNG image.
	png_destroy_read_struct(&png_ptr, &info_ptr, &end_info);
	
	// Successful.
	return 0;
}


void lens_flare_init(void)
{
	// Load the PNG images.
	lens_read_png(lens_flare_320_224, sizeof(lens_flare_320_224), 320, 224, lfraw_320_224);
	lens_read_png(lens_flare_256_224, sizeof(lens_flare_256_224), 256, 224, lfraw_256_224);
	
	// Initialized.
	lf_init = true;
}


void lens_flare_end(void)
{
	lf_init = false;
}


template<typename pixel,
	 unsigned int Rmask, unsigned int Gmask, unsigned int Bmask,
	 unsigned int Rbits, unsigned int Gbits, unsigned int Bbits>
static inline void T_lens_flare_apply(const unsigned int w, const unsigned int h,
				      const uint8_t *src, pixel *screen)
{
	unsigned int x, y;
	unsigned int r, g, b;
	float a;
	
	const unsigned int pitch_diff = (336 - w);
	
	for (y = h; y != 0; y--)
	{
		for (x = w; x != 0; x--)
		{
			if (src[3] == 0)
				goto next_pixel;
			
			if (src[3] == 0xFF)
			{
				// Fully opaque.
				r = src[0];
				g = src[1];
				b = src[2];
			}
			else
			{
				// Translucent.
				b = (*screen & Bmask) << (8 - Bbits);
				g = ((*screen >> Bbits) & Gmask) << (8 - Gbits);
				r = ((*screen >> (Bbits + Gbits)) & Rmask) << (8 - Rbits);
				
				a = (255.0 / (255.0 - src[3]));
				
				r += (src[0] * a);
				g += (src[1] * a);
				b += (src[2] * a);
				
				if (r > 255)
					r = 255;
				if (g > 255)
					g = 255;
				if (b > 255)
					b = 255;
			}
			
			*screen = (b >> (8 - Bbits)) |
				  ((g >> (8 - Gbits)) << Bbits) |
				  ((r >> (8 - Rbits)) << (Bbits + Gbits));
			
next_pixel:
			src += 4;
			screen++;
		}
		
		// Next line.
		screen += pitch_diff;
	}
}

void lens_flare_apply(void)
{
	if (!lf_init)
		lens_flare_init();
	
	const unsigned int w = isFullXRes() ? 320 : 256;
	const uint8_t *src = isFullXRes() ? lfraw_320_224 : lfraw_256_224;
	
	// Lens flare only works for 15/16-bit color.
	// No big deal, since the hq4x renderer doesn't support 32-bit color anyway.
	// NOTE: hq4x doesn't seem to work properly in 15-bit color. (Green artifacting shows up.)
	if (bppMD == 16)
		T_lens_flare_apply<uint16_t, 0x1F, 0x3F, 0x1F, 5, 6, 5>(w, 224, src, &MD_Screen[8]);
	else if (bppMD == 15)
		T_lens_flare_apply<uint16_t, 0x1F, 0x1F, 0x1F, 5, 5, 5>(w, 224, src, &MD_Screen[8]);
}
